from __future__ import annotations

import threading
from collections import deque

from .buffer_input import Input


class TryInput(Input):
    '''
    # TryInput

    A buffered input that checks for type.

    It inherits from `Input`.
    '''

    def __init__(self):
        super().__init__()

    def __call__(self, prompt: str = "", *, cast: type | None = None, default=None):
        '''
        The `try_input()` reads the user input like `buffer_input()`.

        Call `try_input.get_input_buffer()` to get the current buffer.

        Additionally, if `cast` is provided, it will try to cast the result into that type and if
        it fails it will returm `default` (or None in case it wasn't provided).

        ## Arguments

        prompt: str [optional] = what to print before asking for input. default value = "".

        named arguments:
        * cast: type | None [optional] = what to cast the input value to. default = None.
        * default: Any | None [optional] = what to return in case the cast fails. default = None.

        string: str | Any | None [return] = returns a string with the user input or, if it was
        provided, the user input casted into `cast`. if the casting fails it returns `default`.

        ## Usage

        e.g.
        ```python
          number = try_input("Enter a number: ", cast=int, default=0)
        ```
        '''

        if cast is None:
            return super().__call__(prompt)

        try:
            return cast(super().__call__(prompt))
        except ValueError:
            return default


class ThreadSafeInput(Input):
    _input_lock = threading.Lock()
    _history = []
    _buffer = deque()
    _right_buffer = deque()

    def __init__(self):
        super().__init__()

    def __call__(self, prompt: str = ""):
        '''
        The `safe_input()` reads the user input like `buffer_input()` but providing (or attempting)
        thread safety.

        Call `safe_input.get_input_buffer()` to get the current buffer. This buffer is not shared
        between other (non thread safe) input types.

        ## Arguments

        prompt: str [optional] = what to print before asking for input. default value = "".

        string: str | Any | None [return] = returns a string with the user input or, if it was
        provided, the user input casted into `cast`. if the casting fails it returns `default`.

        ## Usage

        e.g.
        ```python
          value = safe_input("Enter a number: ")
        ```
        '''

        with self._input_lock:
            return super().__call__(prompt)


class ThreadSafeTryInput(TryInput):
    _input_lock = ThreadSafeInput._input_lock
    _history = ThreadSafeInput._history
    _buffer = ThreadSafeInput._buffer
    _right_buffer = ThreadSafeInput._right_buffer

    def __init__(self):
        super().__init__()

    def __call__(self, prompt: str = "", *, cast: type | None = None, default=None):
        '''
        The `safe_input()` reads the user input like `buffer_input()` but providing (or attempting)
        thread safety.

        Call `safe_input.get_input_buffer()` to get the current buffer. This buffer is not shared
        between other (non thread safe) input types.

        Additionally, if `cast` is provided, it will try to cast the result into that type and if
        it fails it will returm `default` (or None in case it wasn't provided).

        ## Arguments

        prompt: str [optional] = what to print before asking for input. default value = "".

        named arguments:
        * cast: type | None [optional] = what to cast the input value to. default = None.
        * default: Any | None [optional] = what to return in case the cast fails. default = None.

        string: str | Any | None [return] = returns a string with the user input or, if it was
        provided, the user input casted into `cast`. if the casting fails it returns `default`.

        ## Usage

        e.g.
        ```python
          number = safe_try_input("Enter a number: ", cast=int, default=0)
        ```
        '''

        with self._input_lock:
            return super().__call__(prompt, cast=cast, default=default)


try_input = TryInput()
safe_input = ThreadSafeInput()
safe_try_input = ThreadSafeTryInput()



# getwch code based on code from jfktrey on GitHub: https://gist.github.com/jfktrey/8928865
# try_read code based on this answer on stackoverflow: https://stackoverflow.com/a/5047058
import platform


if platform.system() == "Windows":
    import msvcrt

    def getwch() -> str:
        '''
        Cross platform version of `getwch()`.

        Wide variant of `getch()`, returning a Unicode value.
        '''

        return msvcrt.getwch()

    def try_read() -> str:
        '''
        Tries to read the next character in the stdin buffer and return it. If there is none then it
        returns an empty string.
        '''
        try:
            msvcrt.ungetwch("\n")
            msvcrt.getwch()
            return ""
        except OSError:
            return msvcrt.getwch()
else:
    import tty
    import termios
    import sys
    import fcntl
    import os

    def getwch() -> str:
        '''
        Cross platform version of `getwch()`.

        Wide variant of `getch()`, returning a Unicode value.
        '''

        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)

        try:
            tty.setraw(fd)
            char = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)

        return char

    def try_read() -> str:
        '''
        Tries to read the next character in the stdin buffer and return it. If there is none then it
        returns an empty string.
        '''
        fd = sys.stdin.fileno()
        old_attr = termios.tcgetattr(fd)
        old_flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, old_flags | os.O_NONBLOCK)

        try:
            tty.setraw(fd)
            sys.stdin.read(1)
            char = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_attr)
            fcntl.fcntl(fd, fcntl.F_SETFL, old_flags)

        return char



'''
# Input Handler Module

The `inputhandler` module from the `pyinputhandler` package.

It contains various buffered inputs.

## Usage

You can import the simple `buffer_input()` and use it just like a normal input.

e.g.
```python
  from inputhandler import buffer_input


  value = buffer_input("Enter a number: ")
```

You can also use it in an asynchronous operation and read its buffer with `get_input_buffer()`.

e.g.
```python
  ...  # your async code
  buffer = get_input_buffer()
```
'''

from __future__ import annotations

import threading

from collections import deque

from .stdin import getwch, try_read
from .buffer_input import Input, buffer_input
from .input_types import try_input, safe_input, safe_try_input
from . import input_types
from . import constant


def get_input_buffer() -> deque[str]:
    '''
    An alternative to get the shared buffer input. Returns a buffer as a string deque.
    '''

    return buffer_input.get_input_buffer()


def get_safe_buffer() -> deque[str]:
    '''
    An alternative to get the safe buffer input. Returns a buffer as a string deque.
    '''

    return safe_input.get_input_buffer()


def set_lock(lock: threading.Lock):
    '''
    Sets the `safe_input()` and `safe_try_input()` locks with the provided `lock`.
    '''

    if not isinstance(lock, threading.Lock):
        raise TypeError("lock must be a threading.Lock instance")

    safe_input._input_lock = lock
    safe_try_input._input_lock = lock
    input_types.ThreadSafeInput._input_lock = lock
    input_types.ThreadSafeTryInput._input_lock = lock


def fix_cursor(input_: Input = None):
    '''
    Fixes the cursor position on the console.
    '''

    if isinstance(input_, Input):
        input_._move_cursor()
    elif isinstance(constant.CURRENT_INPUT, Input) or constant.CURRENT_INPUT is None:

        if constant.CURRENT_INPUT is not None:
            constant.CURRENT_INPUT._move_cursor()
    else:
        raise TypeError("the global constant 'CURRENT_INPUT' was reassigned")



from __future__ import annotations

import platform

from collections import deque
from enum import Enum, unique

from .stdin import getwch, try_read
from . import constant

if platform.system() == "Windows":
    # Code to move the cursor on Windows inspired by colorama by tartley
    # GitHub: https://github.com/tartley/colorama/blob/master/colorama/win32.py
    from ctypes import windll, wintypes, Structure, byref

    class ConsoleInfo(Structure):
        _fields_ = [("dwSize", wintypes._COORD), ("dwCursorPosition", wintypes._COORD),
                    ("wAttributes", wintypes.WORD), ("srWindow", wintypes.SMALL_RECT),
                    ("dwMaximumWindowSize", wintypes._COORD)]


@unique
class Key(Enum):
    LEFT = 1
    RIGHT = 2
    UP = 3
    DOWN = 4


class Input:
    '''
    # Input

    A buffered input class.

    It serves as the base class for all the other input classes.
    '''

    _history: list[str] = []
    _buffer: deque[str] = deque()
    _right_buffer: deque[str] = deque()
    _system = platform.system()

    def __init__(self):
        if self._system == "Windows":
            self._console_info = ConsoleInfo()
            self._handle = windll.kernel32.GetStdHandle(-11)
            self._directions = {"K": Key.LEFT, "M": Key.RIGHT, "H": Key.UP, "P": Key.DOWN}
        else:
            self._directions = {"D": Key.LEFT, "C": Key.RIGHT, "A": Key.UP, "B": Key.DOWN}

        self._cursor_left = 0
        self._search_index = 0

    def __call__(self, prompt: str = "") -> str:
        '''
        The `buffer_input()` reads the user input and saves it in a (shared between input types)
        buffer while the user is typing.

        Call `buffer_input.get_input_buffer()` to get the current buffer.

        ## Arguments

        prompt: str [optional] = what to print before asking for input. default value = "".

        string: str [return] = returns a string with the user typed input.

        ## Usage

        e.g.
        ```python
          value = buffer_input("Enter a number: ")
        ```
        '''

        if prompt != "":
            print(prompt, end="", flush=True)

        while True:
            if constant.CURRENT_INPUT is not self:
                constant.CURRENT_INPUT = self

            char = getwch()

            if char == "\r":
                break
            elif char == "\x03":
                self._buffer.clear()
                raise KeyboardInterrupt

            if self._system == "Windows":
                if char == "\x08":
                    if len(self._buffer) > 0:
                        self._buffer.pop()
                        len_ = len(self._buffer) + len(self._right_buffer) + len(prompt) + 1
                        out = ("\r" + " "*len_ + "\r" + prompt
                               + "".join(self._buffer + self._right_buffer))
                        print(out, end="", flush=True)
                        self._move_cursor()
                elif char == "\x00" or char == "à":
                    char = try_read() if char == "à" else getwch()

                    if not char:
                        self._buffer.append("à")
                        print("à" + "".join(self._right_buffer), end="", flush=True)
                        self._move_cursor()
                        continue

                    direction = self._directions.get(char, None)

                    if direction == Key.LEFT:
                        if len(self._buffer) + len(self._right_buffer) == self._cursor_left:
                            continue

                        windll.kernel32.GetConsoleScreenBufferInfo(self._handle,
                                                                   byref(self._console_info))
                        pos = self._console_info.dwCursorPosition
                        new = wintypes._COORD(pos.X - 1, pos.Y)
                        windll.kernel32.SetConsoleCursorPosition(self._handle, new)
                        self._cursor_left += 1
                        self._right_buffer.appendleft(self._buffer.pop())
                    elif direction == Key.RIGHT:
                        if self._cursor_left == 0:
                            continue

                        windll.kernel32.GetConsoleScreenBufferInfo(self._handle,
                                                                   byref(self._console_info))
                        pos = self._console_info.dwCursorPosition
                        new = wintypes._COORD(pos.X + 1, pos.Y)
                        windll.kernel32.SetConsoleCursorPosition(self._handle, new)
                        self._cursor_left -= 1
                        self._buffer.append(self._right_buffer.popleft())
                    elif direction == Key.UP:
                        if len(self._history) == self._search_index * -1:
                            continue

                        self._search_index -= 1
                        len_ = len(self._buffer) + len(self._right_buffer) + len(prompt) + 1
                        self._buffer.clear()
                        self._right_buffer.clear()
                        out = "\r" + " "*len_ + "\r" + prompt + self._history[self._search_index]
                        print(out, end="", flush=True)
                        self._buffer += deque(self._history[self._search_index])
                    elif direction == Key.DOWN:
                        if self._search_index < 0:
                            self._search_index += 1

                        if self._search_index == 0:
                            continue

                        len_ = len(self._buffer) + len(self._right_buffer) + len(prompt) + 1
                        self._buffer.clear()
                        self._right_buffer.clear()
                        out = "\r" + " "*len_ + "\r" + prompt + self._history[self._search_index]
                        print(out, end="", flush=True)
                        self._buffer += deque(self._history[self._search_index])
                    elif direction is None:
                        continue
                else:
                    self._buffer.append(char)
                    print(char + "".join(self._right_buffer), end="", flush=True)
                    self._move_cursor()
            else:
                if char == "\x7f":
                    if len(self._buffer) > 0:
                        self._buffer.pop()
                        len_ = len(self._buffer) + len(self._right_buffer) + len(prompt) + 1
                        out = ("\r" + " "*len_ + "\r" + prompt
                               + "".join(self._buffer + self._right_buffer))
                        out += self._move_cursor()
                        print(out, end="", flush=True)
                elif char == "\x1b":
                    char = try_read()

                    if not char:
                        continue

                    direction = self._directions.get(char, None)

                    if direction == Key.LEFT:
                        if len(self._buffer) + len(self._right_buffer) == self._cursor_left:
                            continue

                        print("\033[D", end="", flush=True)
                        self._cursor_left += 1
                        self._right_buffer.appendleft(self._buffer.pop())
                    elif direction == Key.RIGHT:
                        if self._cursor_left == 0:
                            continue

                        print("\033[C", end="", flush=True)
                        self._cursor_left -= 1
                        self._buffer.append(self._right_buffer.popleft())
                    elif direction == Key.UP:
                        if len(self._history) == self._search_index * -1:
                            continue

                        self._search_index -= 1
                        len_ = len(self._buffer) + len(self._right_buffer) + len(prompt) + 1
                        self._buffer.clear()
                        self._right_buffer.clear()
                        out = "\r" + " "*len_ + "\r" + prompt + self._history[self._search_index]
                        print(out, end="", flush=True)
                        self._buffer += deque(self._history[self._search_index])
                    elif direction == Key.DOWN:
                        if self._search_index < 0:
                            self._search_index += 1

                        if self._search_index == 0:
                            continue

                        len_ = len(self._buffer) + len(self._right_buffer) + len(prompt) + 1
                        self._buffer.clear()
                        self._right_buffer.clear()
                        out = "\r" + " "*len_ + "\r" + prompt + self._history[self._search_index]
                        print(out, end="", flush=True)
                        self._buffer += deque(self._history[self._search_index])
                    elif direction is None:
                        continue
                else:
                    self._buffer.append(char)
                    print(char + "".join(self._right_buffer) + self._move_cursor(), end="",
                          flush=True)

        print()
        string = ""

        while len(self._buffer) > 0:
            string += self._buffer.popleft()

        while len(self._right_buffer) > 0:
            string += self._right_buffer.popleft()

        self._history.append(string)
        self._cursor_left = 0
        self._search_index = 0
        return string

    def _move_cursor(self) -> str:
        if self._cursor_left > 0:
            if self._system == "Windows":
                windll.kernel32.GetConsoleScreenBufferInfo(self._handle, byref(self._console_info))
                pos = self._console_info.dwCursorPosition
                new = wintypes._COORD(pos.X - self._cursor_left, pos.Y)
                windll.kernel32.SetConsoleCursorPosition(self._handle, new)
                return ""
            else:
                return "\033[D" * self._cursor_left
        else:
            return ""

    @classmethod
    def get_input_buffer(cls) -> deque[str]:
        '''
        Get the currently saved (shared between input types) buffer.

        Note: the buffer is not shared between thread safe inputs and normal inputs.

        ## Arguments

        buffer: deque[str] [return] = returns a copy of the buffer as a string deque.
        '''

        if not isinstance(constant.CURRENT_INPUT, Input) and constant.CURRENT_INPUT is not None:
            raise TypeError("the global constant 'CURRENT_INPUT' was reassigned")
        elif constant.CURRENT_INPUT is None:
            return cls._buffer.copy() + cls._right_buffer.copy()
        else:
            move = [] if cls._system == "Windows" else [constant.CURRENT_INPUT._move_cursor()]
            return cls._buffer.copy() + cls._right_buffer.copy() + deque(move)


buffer_input = Input()



from setuptools import setup, find_packages


with open("README.md", "r") as readme:
    description = readme.read()

setup(
    name="pyinputhandler",
    version="1.3.0",
    author="DaHunterTime",
    description="A (basic) cross-platform python input handler",
    long_description=description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/DaHunterTime/InputHandler",
    project_urls={
        "Bug Tracker": "https://github.com/DaHunterTime/InputHandler/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.7",
)
